from dataclasses import dataclass

from energytt_platform.bus import Message
from energytt_platform.models.technologies import Technology


@dataclass
class TechnologyUpdate(Message):
    """
    An update to a Technology.
    """
    technology: Technology



import flask
import serpyco
import rapidjson
from abc import abstractmethod
from typing import List, Dict, Any, Optional

from .context import Context
from .flask import FlaskContext
from .endpoints import Endpoint
from .guards import EndpointGuard, bouncer
from .responses import HttpError, BadRequest


# -- Request data ------------------------------------------------------------

class RequestDataProvider(object):
    """
    Provides request parameters.
    """
    @abstractmethod
    def get(self) -> Optional[Dict[str, Any]]:
        raise NotImplementedError


class JsonBodyProvider(RequestDataProvider):
    """
    Reads request data from request body JSON.
    """
    def get(self) -> Optional[Dict[str, Any]]:
        if not flask.request.data:
            return None

        try:
            return rapidjson.loads(flask.request.data.decode('utf8'))
        except rapidjson.JSONDecodeError:
            raise BadRequest('Invalid JSON body provided')


class QueryStringProvider(RequestDataProvider):
    """
    Reads request data from query parameters.
    """
    def get(self) -> Optional[Dict[str, Any]]:
        return dict(flask.request.args)


# -- Orchestrator ------------------------------------------------------------


class RequestOrchestrator(object):
    """
    Orchestrates handling of HTTP requests on behalf of an endpoint.

    Behaves as a Flask endpoint, ie. it is callable without taking any
    parameters, so it plugs into Flask seamlessly.
    """
    def __init__(
            self,
            endpoint: Endpoint,
            data: RequestDataProvider,
            guards: List[EndpointGuard] = None,
    ):
        self.endpoint = endpoint
        self.data = data
        self.guards = guards

    def __call__(self) -> flask.Response:
        """
        Invoked by Flask to handle a HTTP request.
        """
        try:
            return self.invoke_endpoint()
        except HttpError as e:
            return self.handle_http_error(e)
        except Exception as e:
            raise
            return self.handle_exception(e)

    def build_context(self) -> Context:
        """
        Creates a new request context.
        """
        return FlaskContext()

    def invoke_endpoint(self) -> flask.Response:
        """
        TODO
        """
        context = self.build_context()

        if self.guards:
            bouncer.validate(context, self.guards)

        # kwargs passed to Endpoint.handle_request()
        handler_kwargs = {}

        if self.endpoint.requires_context:
            handler_kwargs['context'] = context

        # Deserialize request data (if necessary)
        if self.endpoint.should_parse_request_data:
            # Defaulting to an empty dictionary makes it possible to omit
            # request data for models where all fields are optional
            handler_kwargs['request'] = self.parse_request_data(self.data.get() or {})

        # Invoke endpoint
        response_body = self.endpoint.handle_request(**handler_kwargs)

        # Serialize response object (if necessary)
        if self.endpoint.should_parse_response_object:
            response_body = self.parse_response_object(response_body)
            response_mimetype = 'application/json'
        else:
            # TODO
            response_body = ''
            response_mimetype = 'text/html'

        return flask.Response(
            status=200,
            mimetype=response_mimetype,
            response=response_body,
        )

    def handle_http_error(self, e: HttpError) -> flask.Response:
        """
        TODO
        """
        return flask.Response(
            response=e.msg,
            status=e.status_code,
            mimetype='text/html',
        )

    def handle_exception(self, e: Exception) -> flask.Response:
        """
        TODO
        """
        return flask.Response(
            response='Internal Server Error',
            status=500,
            mimetype='text/html',
        )

    def parse_request_data(self, data: Dict[str, Any]) -> Any:
        """
        TODO
        """
        try:
            return self.endpoint.request_serializer.load(data, True)
        except serpyco.exception.ValidationError as e:
            # JSON schema validation failed for request data
            # TODO Parse ValidationError to something useful
            raise BadRequest(str(e))

    def parse_response_object(self, response: Any) -> str:
        """
        TODO
        """
        return self.endpoint.response_serializer.dump_json(response)



from functools import partial


class HttpResponse(object):
    def __init__(self, status: int, msg: str = ''):
        self.status = status
        self.msg = msg

    @classmethod
    def build(cls, *args, **kwargs):
        return partial(cls, *args, **kwargs)


class HttpError(HttpResponse, Exception):
    def __init__(self, msg: str, status_code: int):
        super(HttpError, self).__init__('%d %s' % (status_code, msg))
        self.msg = msg
        self.status_code = status_code

    @classmethod
    def build(cls, status_code: int):
        return partial(cls, status_code=status_code)


BadRequest = HttpError.build(400)
Unauthorized = HttpError.build(401)
InternalServerError = HttpError.build(500)





class SqlQuery(object):
    """
    TODO
    """
    def __init__(self, session, q):
        """
        :param sqlalchemy.orm.Session session:
        :param sqlalchemy.orm.Query q:
        """
        self.session = session
        self.q = q

    def __iter__(self):
        return iter(self.q)

    def __getattr__(self, name):
        return getattr(self.q, name)

    def filter(self, *filters):
        return self.__class__(self.session, self.q.filter(*filters))

    def unique_join(self, *props, **kwargs):
        if props[0] in [c.entity for c in self.q._join_entities]:
            return self

        return self.__class__(self.session, self.q.join(
            *props, **kwargs
        ))

    def lazy_load(self, field):
        return self.__class__(self.session, self.q.options(
            lazyload(field)
        ))

    def eager_load(self, path, *fields):
        if isinstance(path, (list, tuple)):
            join = joinedload(*path)
        else:
            join = joinedload(path)

        if fields:
            join = join.load_only(*fields)
        return self.__class__(self.session, self.q.options(
            join
        ))

    # def eage_load_path(self, *path):
    #     join = rjoinedload(*path)
    #     if fields:
    #         join = join.load_only(*fields)
    #     return self.__class__(self.session, self.q.options(
    #         join
    #     ))

    def only(self, *fields):
        return self.__class__(self.session, self.q.options(
            load_only(*fields)
        ))

    def get(self, field):
        return self.only(field).scalar()

    def exists(self):
        return self.count() > 0



from typing import List
from dataclasses import dataclass

from energytt_platform.serialize import JsonSerializer, Serializable


@dataclass
class NestedTestSerializable(Serializable):
    friends: List[str]


@dataclass
class TestSerializable(Serializable):
    name: str
    age: int
    nested: NestedTestSerializable


class TestJsonSerializer:

    def test__should_serialize_and_deserialize_correctly(self):

        # -- Arrange ---------------------------------------------------------

        obj = TestSerializable(
            name='John',
            age=50,
            nested=NestedTestSerializable(
                friends=['Bill', 'Joe', 'Tedd'],
            )
        )

        uut = JsonSerializer()

        # -- Act -------------------------------------------------------------

        serialized = uut.serialize(obj)
        deserialized = uut.deserialize(serialized, TestSerializable)

        # -- Assert ----------------------------------------------------------

        assert deserialized == obj



from typing import List

from .registry import message_registry
from .kafka import KafkaMessageBroker
from .serialize import MessageSerializer
from .dispatcher import MessageDispatcher
from .broker import MessageBroker, Message


def get_default_broker(servers: List[str]) -> MessageBroker:
    """
    Creates and returns an instance of the default message broker.

    :param List[str] servers:
    :rtype: MessageBroker
    :return: An instance of the default message broker
    """
    return KafkaMessageBroker(
        servers=servers,
        serializer=MessageSerializer(),
    )



from typing import Dict
from flask import request
from functools import cached_property

from energytt_platform.api.context import Context


class FlaskContext(Context):
    """
    Flask-specific context.
    """

    @cached_property
    def headers(self) -> Dict[str, str]:
        """
        Returns request headers.
        """
        return dict(request.headers)



from typing import Optional
from dataclasses import dataclass, field

from energytt_platform.bus import Message
from energytt_platform.models.common import Address
from energytt_platform.models.technologies import Technology
from energytt_platform.models.meteringpoints import \
    MeteringPoint, MeteringPointType


@dataclass
class MeteringPointAdded(Message):
    """
    A new MeteringPoint has been added to the system.
    """
    subject: str
    meteringpoint: MeteringPoint


@dataclass
class MeteringPointRemoved(Message):
    """
    A MeteringPoint has been remove from the system.
    """
    gsrn: str


@dataclass
class MeteringPointMetaDataUpdate(Message):
    """
    Metadata for a MeteringPoint has been updated.
    """
    gsrn: str
    type: Optional[MeteringPointType] = field(default=None)
    technology: Optional[Technology] = field(default=None)
    address: Optional[Address] = field(default=None)



from energytt_platform.bus import message_registry

from .auth import UserOnboarded
from .measurements import MeasurementAdded
from .meteringpoints import MeteringPointAdded, MeteringPointRemoved


message_registry.add(
    UserOnboarded,
    MeasurementAdded,
    MeteringPointAdded,
    MeteringPointRemoved,
)



from typing import List
from abc import abstractmethod
from dataclasses import dataclass

from energytt_platform.serialize import Serializable

from .context import Context


@dataclass
class Issuer(Serializable):
    id: str
    public_key: str


class EndpointGuard(object):

    @abstractmethod
    def validate(self, context: Context):
        raise NotImplementedError


# class ServiceGuard(EndpointGuard):
#     """
#     Allows only specific services to access this endpoint.
#     """
#     def __init__(self, *services: Service):
#         self.services = services
#
#
# class IssuerGuard(EndpointGuard):
#     """
#     Allows only specific issuers to access this endpoint.
#     """
#     def __init__(self, *services: Service):
#         self.services = services


class ScopedGuard(EndpointGuard):
    """
    Allows only clients with specific scopes granted.
    """
    def __init__(self, *scopes: str):
        self.scopes = scopes


class Bouncer(object):
    def validate(self, context: Context, guards: List[EndpointGuard]):
        raise NotImplementedError


# s1 = ServiceGuard(
#     Service(name='Service A'),
#     Service(name='Service B'),
# )
#
# s2 = ScopedGuard(
#     'meteringpoints.read',
#     'measurements.read',
#     'gc.read',
#     'gc.transfer',
# )


# -- Singletons --------------------------------------------------------------

bouncer = Bouncer()



from serpyco import Serializer
from abc import abstractmethod
from typing import Optional, Any
from inspect import getfullargspec
from functools import cached_property


class Endpoint(object):
    """
    Represents a single HTTP API endpoint.
    """

    # Request and response schemas (dataclasses or None)
    Request = None
    Response = None

    @cached_property
    def request_serializer(self) -> Optional[Serializer]:
        """
        Returns serpyco Serializer for request model.
        """
        return self.build_request_serializer() \
            if self.Request is not None \
            else None

    @cached_property
    def response_serializer(self) -> Optional[Serializer]:
        """
        Returns serpyco Serializer for response model.
        """
        return self.build_response_serializer() \
            if self.Response is not None \
            else None

    def build_request_serializer(self) -> Serializer:
        """
        Builds serpyco Serializer for request model.
        """
        return Serializer(self.Request)

    def build_response_serializer(self) -> Serializer:
        """
        Builds serpyco Serializer for response model.
        """
        return Serializer(self.Response)

    @cached_property
    def requires_context(self) -> bool:
        """
        Returns True if handle_request() requires a Context passed as argument.
        """
        return 'context' in getfullargspec(self.handle_request)[0]

    @cached_property
    def should_parse_request_data(self) -> bool:
        """
        Returns True if handle_request() requires an instance of self.Request
        passed as argument.
        """
        return (self.Request is not None
                and 'request' in getfullargspec(self.handle_request)[0])

    @cached_property
    def should_parse_response_object(self) -> bool:
        """
        Returns True if handle_request() requires an instance of self.Request
        passed as argument.
        """
        return self.Response is not None

    @abstractmethod
    def handle_request(self, **kwargs) -> Optional[Any]:
        """
        Handle a HTTP request.
        """
        raise NotImplementedError


class HealthCheck(Endpoint):
    """
    Health check endpoint. Always returns status 200.
    """
    def handle_request(self):
        pass



from serpyco import field
from dataclasses import dataclass
from typing import Generic, TypeVar, Union, Dict, Any, Optional

from energytt_platform.serialize import json_serializer, simple_serializer

from .broker import Message
from .registry import message_registry


TWrappedMessage = TypeVar('TWrappedMessage', bound=Message)
TSerializedMessage = Dict[str, Any]


@dataclass
class MessageWrapper(Message, Generic[TWrappedMessage]):
    """
    A new MeteringPoint has been added to the system.
    """
    type_: str = field(dict_key='type')
    msg: Optional[Union[TWrappedMessage, TSerializedMessage]] = field(default=None)


class MessageSerializer(object):
    """
    A serializer specifically for serializing messages to
    and from the event bus.
    """

    class MessageSerializeError(Exception):
        pass

    class MessageDeserializeError(Exception):
        pass

    def serialize(self, msg: Message) -> bytes:
        """
        Wraps message appropriately and JSON serializes it.
        """
        if msg not in message_registry:
            raise self.MessageSerializeError((
                f'Can not serialize of type "{msg.__class__.__name__}": '
                'Type is unknown to the bus.'
            ))

        wrapped_msg = MessageWrapper(
            type_=msg.__class__.__name__,
            msg=msg,
        )

        return json_serializer.serialize(wrapped_msg)

    def deserialize(self, data: bytes) -> Message:
        """
        Deserializes JSON bytestream into a message.
        """
        wrapped_msg = json_serializer.deserialize(data, MessageWrapper)

        if wrapped_msg.type_ not in message_registry:
            raise self.MessageDeserializeError((
                f'Can not deserialize message of type "{wrapped_msg.type_}": '
                'Type is unknown to the bus.'
            ))

        message_cls = message_registry.get(wrapped_msg.msg)

        return simple_serializer.deserialize(
            data=wrapped_msg.msg,
            cls=message_cls,
        )

# class MessageSerializerOLD(object):
#     """
#     A serializer specifically for serializing messages to
#     and from the event bus.
#     """
#
#     SEPARATOR = b'\n'
#
#     def __init__(self):
#         self.serializer = JsonSerializer()
#
#     def serialize(self, msg: Serializable) -> bytes:
#         if msg.type_name not in registry:
#             raise RuntimeError((
#                 'Can not send message of type "%s": '
#                 'Type is not known by the bus.'
#             ) % msg.type_name)
#
#         return b'%b%b%b' % (
#             msg.type_name.encode(),
#             self.SEPARATOR,
#             self.serializer.serialize(msg),
#         )
#
#     def deserialize(self, data: bytes) -> Serializable:
#         separator_index = data.find(self.SEPARATOR)
#         object_name = data[:separator_index].decode('utf8')
#         object_class = registry[object_name]
#         object_data = data[separator_index+1:]
#
#         return self.serializer.deserialize(object_data, object_class)



from typing import List, Any
from functools import cached_property
from kafka import KafkaProducer, KafkaConsumer
from kafka.errors import KafkaError

from energytt_platform.bus.broker import MessageBroker
from energytt_platform.bus.exceptions import PublishError
from energytt_platform.bus.serialize import MessageSerializer


class KafkaMessageConsumer(object):
    """
    A consumer of Kafka messages.
    Iterates over messages in subscribed topics.
    """
    def __init__(
            self,
            topics: List[str],
            servers: List[str],
            serializer: MessageSerializer,
    ):
        """
        :param List[str] topics:
        :param List[str] servers:
        :param MessageSerializer serializer:
        """
        self.topics = topics
        self.servers = servers
        self.serializer = serializer

    @cached_property
    def consumer(self) -> KafkaConsumer:
        """
        TODO
        """
        return KafkaConsumer(
            *self.topics,
            bootstrap_servers=self.servers,
            value_deserializer=self.serializer.deserialize,
            auto_offset_reset='earliest',
            enable_auto_commit=False,
        )

    def __iter__(self):
        return (msg.value for msg in self.consumer)


class KafkaMessageBroker(MessageBroker):
    """
    Implementation of Kafka as message bus.
    """
    def __init__(self, servers: List[str], serializer: MessageSerializer):
        self.servers = servers
        self.serializer = serializer

    @cached_property
    def producer(self) -> KafkaProducer:
        """
        TODO
        """
        return KafkaProducer(
            bootstrap_servers=self.servers,
            value_serializer=self.serializer.serialize,
        )

    def publish(self, topic: str, msg: Any, block=True, timeout=10):
        """
        TODO
        """
        future = self.producer.send(
            topic=topic,
            value=msg,
        )

        if block:
            try:
                record_metadata = future.get(timeout=timeout)
            except KafkaError as e:
                # Decide what to do if produce request failed...
                raise PublishError(str(e))

    def listen(self, topics: List[str]) -> KafkaMessageConsumer:
        """
        TODO
        """
        return KafkaMessageConsumer(
            topics=topics,
            servers=self.servers,
            serializer=self.serializer,
        )



from enum import Enum
from typing import Optional
from dataclasses import dataclass, field

from energytt_platform.serialize import Serializable


class TechnologyLabel(Enum):
    """
    System-wide labels of known technologies.
    """
    COAL = 'Coal'
    NUCLEAR = 'Nuclear'
    SOLAR = 'Solar'
    WIND = 'Wind'


@dataclass
class Technology(Serializable):
    """
    A technology described by the standard described in the
    EECS Rules Fact Sheet 5: TYPES OF ENERGY INPUTS AND TECHNOLOGIES
    """
    technology_code: str
    fuel_code: str
    label: Optional[TechnologyLabel] = field(default=None)  # TODO Why optional ???



# from typing import List
# from dataclasses import dataclass
#
# from energytt_platform.serialize import Serializable
# from energytt_platform.models.meteringpoints import MeteringPoint
# from energytt_platform.models.certificates import GranularCertificate
#
#
# @dataclass
# class CertificateIssued(Serializable):
#     """
#     A new Granular Certificate has been issued.
#     """
#     certificate: GranularCertificate
#
#
# @dataclass
# class CertificateTransferred(Serializable):
#     """
#     A Granular Certificate has been transferred to a new owner.
#     """
#     certificate: GranularCertificate
#     child: GranularCertificate
#
#
# @dataclass
# class CertificateSplit(Serializable):
#     """
#     A Granular Certificate has been split into multiple new certificates.
#     """
#     certificate: GranularCertificate
#     children: List[GranularCertificate]
#
#
# @dataclass
# class CertificateRetired(Serializable):
#     """
#     A Granular Certificate has been retired onto a Metering Point.
#     """
#     certificate: GranularCertificate
#     meteringpoint: MeteringPoint



import serpyco
from abc import abstractmethod
from functools import lru_cache
from dataclasses import dataclass
from typing import Dict, Type, TypeVar, Generic, Any


@dataclass
class Serializable:
    """
    Base class for dataclasses that can be serialized and deserialized.
    Subclasses must be defined as dataclasses.
    """
    pass


# -- Interfaces --------------------------------------------------------------


TSerializable = TypeVar('TSerializable', bound=Serializable)
TSerialized = TypeVar('TSerialized')


class Serializer(Generic[TSerialized]):
    """
    An interface for serializing and deserializing dataclasses.
    """

    @abstractmethod
    def serialize(self, obj: TSerializable) -> TSerialized:
        """
        Serialize an object.
        """
        raise NotImplementedError

    @abstractmethod
    def deserialize(self, data: TSerialized, cls: Type[TSerializable]) -> TSerializable:
        """
        Deserialize an object.
        """
        raise NotImplementedError


# -- Serializers -------------------------------------------------------------


class SimpleSerializer(Serializer[Dict[str, Any]]):
    """
    Serialize and deserialize to and from simple Python types (dictionary).
    """
    def serialize(self, obj: TSerializable) -> Dict[str, Any]:
        """
        Serializes object to Python.
        """
        return get_serializer(obj.__class__).dump(obj)

    def deserialize(self, data: Dict[str, Any], cls: Type[TSerializable]) -> TSerializable:
        """
        Deserialize JSON data to instance of type "cls".
        """
        return get_serializer(cls).load(data)


class JsonSerializer(Serializer[bytes]):
    """
    Serialize and deserialize to and from JSON (encoded bytes).
    """
    def serialize(self, obj: TSerializable) -> bytes:
        """
        Serializes object to JSON.
        """
        return get_serializer(obj.__class__).dump_json(obj).encode()

    def deserialize(self, data: bytes, cls: Type[TSerializable]) -> TSerializable:
        """
        Deserialize JSON data to instance of type "cls".
        """
        return get_serializer(cls).load_json(data.decode('utf8'))


# -- Misc --------------------------------------------------------------------


@lru_cache
def get_serializer(cls: Type[TSerializable]) -> serpyco.Serializer:
    """
    TODO
    """
    return serpyco.Serializer(cls)


# -- Singletons --------------------------------------------------------------


json_serializer = JsonSerializer()
simple_serializer = SimpleSerializer()



from dataclasses import dataclass

from energytt_platform.bus import Message


# @dataclass
# class UserCreated(Serializable):
#     pass
#
#
# @dataclass
# class UserProfileUpdated(Serializable):
#     pass
#
#
# @dataclass
# class UserDeleted(Serializable):
#     pass
#
#
# @dataclass
# class UserConsentGiven(Serializable):
#     pass
#
#
# @dataclass
# class UserConsentRevoked(Serializable):
#     pass


@dataclass
class UserOnboarded(Message):
    """
    A new user has been onboarded to the system.
    """
    subject: str
    name: str



from abc import abstractmethod
from dataclasses import dataclass
from typing import List, Iterable, Callable, Any

from energytt_platform.serialize import Serializable


# TMessage = Serializable


@dataclass
class Message(Serializable):
    """
    TODO
    """
    pass


TMessageHandler = Callable[[Message], None]


class MessageBroker(object):
    """
    Abstract base-class for publishing and consuming messages
    on the message-bus.
    """

    @abstractmethod
    def publish(self, topic: str, msg: Any, block=False, timeout=10):
        """
        Publish a message to a topic on the bus.

        :param str topic: The topic to publish to
        :param Any msg: The message to publish
        :param bool block: Whether to block until publishing is complete
        :param int timeout: Timeout in seconds (if block=True)
        """
        raise NotImplementedError

    @abstractmethod
    def listen(self, topics: List[str]) -> Iterable[Message]:
        """
        Subscribe to one or more topics. Returns an iterable of messages.

        :param List[str] topics: The topics to subscribe to
        :rtype: Iterable[Any]
        :return: An iterable of messages
        """
        raise NotImplementedError

    def subscribe(self, topics: List[str], handler: TMessageHandler):
        """
        An alias for subscribe() except this function takes a callable
        which is invoked for each message.

        :param List[str] topics: The topics to subscribe to
        :param TMessageHandler handler: Message handler
        """
        for msg in self.listen(topics):
            handler(msg)



from datetime import datetime

from energytt_platform.bus import messages as m
from energytt_platform.bus.serialize import MessageSerializer
from energytt_platform.models.measurements import \
    Measurement, MeasurementType


class TestMessageSerializer:

    def test__should_serialize_and_deserialize_correctly(self):

        # -- Arrange ---------------------------------------------------------

        obj = m.MeasurementAdded(
            subject='12345',
            measurement=Measurement(
                type=MeasurementType.CONSUMPTION,
                gsrn='12345',
                amount=123,
                begin=datetime.now(),
                end=datetime.now(),
            )
        )

        uut = MessageSerializer()

        # -- Act -------------------------------------------------------------

        serialized = uut.serialize(obj)
        deserialized = uut.deserialize(serialized)

        # -- Assert ----------------------------------------------------------

        assert isinstance(deserialized, m.MeasurementAdded)
        assert deserialized == obj



from datetime import datetime
from typing import List, Optional
from dataclasses import dataclass

from energytt_platform.serialize import Serializable


@dataclass
class GarbageToken(Serializable):
    issued: datetime
    expires: datetime
    subject: str
    scope: List[str]
    on_behalf_of: Optional[str]

    @property
    def actual_subject(self) -> str:
        if self.on_behalf_of:
            return self.on_behalf_of
        else:
            return self.subject


@dataclass
class OpaqueToken(Serializable):
    issued: datetime
    expires: datetime
    subject: str
    scope: List[str]
    on_behalf_of: Optional[str]

    @property
    def actual_subject(self) -> str:
        if self.on_behalf_of:
            return self.on_behalf_of
        else:
            return self.subject


def encode_opaque_token(token: OpaqueToken) -> str:
    pass


def decode_opaque_token(encoded_jwt: str) -> OpaqueToken:
    pass



from dataclasses import dataclass

from energytt_platform.bus import Message
from energytt_platform.models.measurements import Measurement


@dataclass
class MeasurementAdded(Message):
    """
    A new Measurement has been added to the system.
    """
    subject: str
    measurement: Measurement



import re
import jwt
from abc import abstractmethod
from typing import Dict, Optional
from functools import cached_property

from energytt_platform.serialize import json_serializer
from energytt_platform.auth import (
    OpaqueToken,
    encode_opaque_token,
    decode_opaque_token,
)


class Context(object):
    """
    Context for a single incoming HTTP request.
    """

    TOKEN_HEADER = 'Authorization'
    TOKEN_PATTERN = re.compile(r'^Bearer:\s*(.+)$')

    @property
    @abstractmethod
    def headers(self) -> Dict[str, str]:
        """
        Returns request headers.
        """
        raise NotImplementedError

    @cached_property
    def raw_token(self) -> Optional[str]:
        """
        Returns request Bearer token.
        """
        if self.TOKEN_HEADER in self.headers:
            matches = self.TOKEN_PATTERN.findall(self.headers[self.TOKEN_HEADER])
            if matches:
                return matches[0]

    @cached_property
    def token(self) -> Optional[OpaqueToken]:
        """
        Parses token into an OpaqueToken.
        """
        if self.raw_token is not None:
            return decode_opaque_token(self.raw_token)
